import inspect
import typing

import numba
from numba import njit
from numba.experimental import jitclass as _jitclass
from numba.experimental.jitclass.base import JitClassType

_annotations = {}


def jitclass(cls: typing.Type) -> typing.Type:
    if numba.config.DISABLE_JIT:
        return cls

    # print()
    # print(cls)

    try:
        annos = cls.__annotations__
    except AttributeError:
        spec = {}
    else:
        spec = dict(annos)
        del cls.__annotations__
        _annotations[cls.__qualname__] = spec

    values = {}
    for name, typ in spec.items():
        try:
            value = getattr(cls, name)
        except AttributeError:
            pass
        else:
            values[name] = value
            delattr(cls, name)

        if isinstance(typ, JitClassType):
            spec[name] = typ.class_type.instance_type

    if values:
        sep = "\n "

        body = ""
        init_locals = {}

        for name, value in values.items():
            newname = f"__init__local__{name}"
            body += sep + f"self.{name} = {newname}"
            init_locals[newname] = value

        if isinstance(cls.__init__, type(object.__init__)):
            cls__init__sig = "(self)"
        else:
            init_locals["__cls__init__"] = njit(cls.__init__)

            signature = inspect.signature(cls.__init__)
            param_names = ", ".join(signature.parameters)
            cls__init__sig = f"({param_names})"

            body += sep + f"__cls__init__{cls__init__sig}"

        code = f"def __jitcls__init__{cls__init__sig}:{body}"

        # print(code)
        # print()

        exec(code, init_locals)

        cls.__init__ = init_locals["__jitcls__init__"]

    return _jitclass(spec)(cls)



import numba as nb
import numpy as np

from suchak.jitclass import jitclass


@jitclass
class Window:
    offset: nb.int32
    period: nb.int32

    _buf: nb.double[:]
    _idx: nb.int32 = 0

    def __init__(self, period: int):
        self.offset = period - 1
        self.period = period

        self._buf = np.empty(period)
        self._buf[:] = np.nan

    def next(self, x: float) -> np.ndarray:
        self._buf[self._idx] = x
        self._idx = (self._idx + 1) % self.period
        return self._buf



import numba as nb
import numpy as np

from suchak.sma import SMA
from suchak.jitclass import jitclass


@jitclass
class EMA:
    offset: nb.int32
    period: nb.int32

    _ema: nb.double = np.nan
    _sma: SMA
    _alpha: nb.double = np.nan

    def __init__(self, period: int):
        self.period = period

        self._ema = np.nan
        self._sma = SMA(period)
        self._alpha = 2 / (period + 1)
        self.offset = self._sma.offset

    def next(self, x: float) -> float:
        if np.isnan(self._ema):
            self._ema = self._sma.next(x)
        self._ema = self._alpha * x + (1 - self._alpha) * self._ema
        return self._ema



import numba as nb

from suchak.ema import EMA
from suchak.jitclass import jitclass
import numpy as np


@jitclass
class RSI:
    offset: nb.int32
    period: nb.int32

    _rma_up: EMA
    _rma_dn: EMA
    _c1: nb.double = np.nan

    def __init__(self, period: int = 14):
        self.period = period

        self._rma_up = EMA(period)
        self._rma_up._alpha = 1 / period

        self._rma_dn = EMA(period)
        self._rma_dn._alpha = 1 / period

        self.offset = self._rma_up.offset

    def next(self, c: float):
        up = max(c - self._c1, 0)
        dn = max(self._c1 - c, 0)

        dn_next = self._rma_dn.next(dn)
        if dn_next == 0:
            rs = np.nan
        else:
            rs = self._rma_up.next(up) / dn_next
        res = rs / (1 + rs) * 100

        self._c1 = c

        return res



import numba as nb

from suchak.sma import SMA
from suchak.tr import TR
from suchak.jitclass import jitclass


@jitclass
class ATR:
    offset: nb.int32
    period: nb.int32

    _sma: SMA
    _tr: TR

    def __init__(self, period: int):
        self.period = period

        self._sma = SMA(period)
        self._tr = TR()

        self.offset = self._sma.offset + self._tr.offset

    def next(self, h: float, l: float, c: float) -> float:
        return self._sma.next(self._tr.next(h, l, c))



import numba as nb
import numpy as np

from suchak import Window
from suchak.jitclass import jitclass


@jitclass
class Fibo:
    offset: nb.int32

    period: nb.int32

    _h_win: Window
    _l_win: Window

    def __init__(self, period: int = 144):
        self.offset = period - 1
        self.period = period

        self._h_win = Window(period)
        self._l_win = Window(period)

    def next(self, h: float, l: float) -> np.ndarray:
        h_arr = self._h_win.next(h)
        l_arr = self._l_win.next(l)

        h1 = np.max(h_arr)
        l1 = np.min(l_arr)
        fark = h1 - l1

        fark236 = fark * 0.236
        fark382 = fark * 0.382
        fark500 = fark * 0.500
        fark618 = fark * 0.618
        fark786 = fark * 0.786

        hl236 = l1 + fark236
        hl382 = l1 + fark382
        hl500 = l1 + fark500
        hl618 = l1 + fark618
        hl786 = l1 + fark786

        lh236 = h1 - fark236
        lh382 = h1 - fark382
        lh500 = h1 - fark500
        lh618 = h1 - fark618
        lh786 = h1 - fark786

        hbars = np.argmax(h_arr)
        lbars = np.argmin(l_arr)

        # pick the farthest one from last (lower idx)
        cond = hbars > lbars
        f236 = hl236 if cond else lh236
        f382 = hl382 if cond else lh382
        f500 = hl500 if cond else lh500
        f618 = hl618 if cond else lh618
        f786 = hl786 if cond else lh786

        return np.array([l1, h1, f236, f382, f500, f618, f786])



import numba as nb
import numpy as np

from suchak.jitclass import jitclass


@jitclass
class TR:
    offset: nb.int32 = 0

    _c1: nb.double = np.nan

    def next(self, h: float, l: float, c: float) -> float:
        if np.isnan(self._c1):
            ret = h - l
        else:
            ret = max(h - l, abs(h - self._c1), abs(l - self._c1))
        self._c1 = c
        return ret



import numba as nb
import numpy as np

from suchak.jitclass import jitclass
from suchak.window import Window


@jitclass
class SMA:
    offset: nb.int32
    period: nb.int32

    _win: Window

    def __init__(self, period: int):
        self.period = period

        self._win = Window(period)
        self.offset = self._win.offset

    def next(self, x: float) -> float:
        return np.sum(self._win.next(x)) / self.period



from .window import Window
from .atr import ATR
from .deque import Deque
from .ema import EMA
from .fibo import Fibo
from .macd import MACD
from .rsi import RSI
from .sma import SMA
from .supertrend import Supertrend
from .tr import TR



import typing

import numba as nb

from suchak.ema import EMA
from suchak.sma import SMA
from suchak.jitclass import jitclass


@jitclass
class MACD:
    offset: nb.int32
    short_period: nb.int32
    long_period: nb.int32
    signal_period: nb.int32

    _short_ema: EMA
    _long_ema: EMA
    _signal_sma: SMA

    def __init__(
        self, short_period: int = 12, long_period: int = 26, signal_period: int = 9
    ):
        self.short_period = short_period
        self.long_period = long_period
        self.signal_period = signal_period

        self._short_ema = EMA(short_period)
        self._long_ema = EMA(long_period)
        self._signal_sma = SMA(signal_period)

        self.offset = self._signal_sma.offset + max(
            self._short_ema.offset, self._long_ema.offset
        )

    def next(self, c: float) -> typing.Tuple[float, float, float]:
        macd = self._short_ema.next(c) - self._long_ema.next(c)
        signal = self._signal_sma.next(macd)
        hist = macd - signal
        return macd, signal, hist



import typing

import numba as nb
import numpy as np

from suchak.atr import ATR
from suchak.jitclass import jitclass


@jitclass
class Supertrend:
    offset: nb.int32
    period: nb.int32
    factor: nb.double = np.nan

    _atr: ATR

    _dt: nb.double = 1
    _up1: nb.double = np.nan
    _dn1: nb.double = np.nan
    _c1: nb.double = np.nan

    def __init__(self, period: int = 5, factor: float = 3):
        self.period = period
        self.factor = factor

        self._atr = ATR(period)
        self.offset = self._atr.offset

    def next(self, h: float, l: float, c: float) -> typing.Tuple[float, float]:
        f_atr = self.factor * self._atr.next(h, l, c)

        hl2 = (h + l) / 2

        up = hl2 - f_atr
        dn = hl2 + f_atr

        if self._c1 > self._up1 > up:
            up = self._up1
        if self._c1 < self._dn1 < dn:
            dn = self._dn1

        if (self._dt < 0 and self._dn1 < c) or (self._dt > 0 and self._up1 > c):
            self._dt *= -1

        if self._dt > 0:
            st = up
        else:
            st = dn

        self._up1 = up
        self._dn1 = dn
        self._c1 = c

        return st, self._dt



import numba as nb
import numpy as np

from suchak.jitclass import jitclass


@jitclass
class Deque:
    offset: nb.int32
    maxlen: nb.int32

    _buf: nb.double[:]
    _idx0: nb.int32
    len: nb.int32 = 0

    def __init__(self, maxlen: int):
        self.offset = maxlen - 1
        self.maxlen = maxlen

        self._buf = np.empty(maxlen * 4)
        self._idx0 = len(self._buf) // 2

    def append(self, x: float):
        real_len = self._idx0 + self.len

        if real_len >= len(self._buf):
            self._move()

            # need to re-compute this
            real_len = self._idx0 + self.len

        self._buf[real_len] = x

        if self.len < self.maxlen:
            self.len += 1
        else:
            self._idx0 += 1

    def append_left(self, x: float):
        if self._idx0 <= 0:
            self._move()

        self._buf[self._idx0 - 1] = x

        self._idx0 -= 1
        if self.len < self.maxlen:
            self.len += 1

    def _move(self):
        arr = self._buf[self._idx0 : self._idx0 + self.len]
        self._idx0 = len(self._buf) // 2
        self._buf[self._idx0 : self._idx0 + self.len] = arr

    def array(self) -> np.ndarray:
        return self._buf[self._idx0 : self._idx0 + self.len]

    def __getitem__(self, idx: int) -> float:
        return self.array()[idx]
        # return self._buf[(self._idx0 + idx) % self.maxlen]

    def __setitem__(self, idx: int, item: float):
        self.array()[idx] = item
        # self._buf[(self._idx0 + idx) % self.maxlen] = item
